/************************************************************************/
/*																		*/
/*		trace_monitor - A/ROSE Trace Monitor							*/
/*																		*/
/*		Richard K. Nordin			11/20/87.							*/
/*																		*/
/*		Copyright  1987-88, Apple Computer, Inc.  All rights reserved.	*/
/*																		*/
/************************************************************************/


/*	TraceMonitor - Trace Monitor Program
 *
 *	TraceMonitor receives messages from Trace Managers and records them on a
 *	trace file.
 *
 *	TraceMonitor [file]
 *
 *	file	- File Name
 *
 *	If no file name is specified, TraceMonitor records on "TraceFile".
 *
 *	A trace file consists of a sequence of A/ROSE messages.  If a message has
 *	an associated data buffer (mDataSize is nonzero), the message is immediately
 *	followed by the data buffer contents of size mDataSize.
 *
 */

 
#include <StdIO.h>
#include <Types.h>
#include <CType.h>
#include <Desk.h>
#include <Strings.h>
#include <QuickDraw.h>

#include "os.h"
#include "managers.h"


#define TRUE	1
#define	FALSE	0

#define TRACEFILE	"TraceFile"		/* Default Trace File Name */

#define ERRORCODE(x)		((x) |= 0x8000)
#define PREVERROR(x)		((x) &  0x8000)

pascal void
illegal()
	extern	0x4afc;

static	void		Wait_rcv();


Boolean			Q_open = FALSE;		/* IPC Queue Open Flag */


main(argc, argv, envp)
int		argc;			/* Argument Count */
char	*argv[];		/* Arguments */
char	*envp[];		/* Environment */
{

#pragma	unused (envp)

	char			*copy_data;								/* Pointer to Copy of Data */
	FILE			*file;									/* Trace File Descriptor */
	long			space;									/* Current Space Available */
	struct mMessage	copy_msg;								/* Copy of Message */
	struct mMessage	*msg;									/* Pointer to Message */
	void			clean_up();								/* Clean Up */
	static char		my_object_name[] = "Trace Monitor";		/* Trace Monitor Object Name */
	static char		my_type_name[] = "Trace Monitor";		/* Trace Monitor Type Name */
	tid_type		my_tid;									/* TID of Trace Monitor */
	void			send_trace_managers();					/* Send Message to Trace Managers */
	void			record_message();						/* Record Trace Message */

	
	InitGraf( &qd.thePort );
	
	/* Trap exits. */
	
	atexit (clean_up);
	
	/* Process arguments. */
	
	if (argc == 1)
	{
		file = fopen(TRACEFILE, "w");
		if (file == NULL)
		{
			fprintf(stderr, "### TraceMonitor: unable to open %s\n", TRACEFILE);
			exit(1);
		}
	}
	
	if (argc == 2)
	{
		file = fopen(argv[1], "w");
		if (file == NULL)
		{
			fprintf(stderr, "### TraceMonitor: unable to open %s\n", argv[1]);
			exit(1);
		}
	}
	
	if (argc > 2)
	{
		fprintf(stderr, "### TraceMonitor: too many arguments\n");
		fprintf(stderr, "# Usage: TraceMonitor [file]\n");
		exit(1);
	}
	
	/* Get some space for copy of message data. */
	
	space = 32;
	copy_data = (char *) malloc(space);
	if (copy_data == 0)
	{
		fprintf(stderr, "### TraceMonitor: out of memory\n");
		exit(2);
	}

	/* Open IPC queue. */
	
	if (!OpenQueue(Wait_rcv))
	{
		fprintf(stderr, "### TraceMonitor: unable to open queue\n");
		exit(2);
	}
	
	Q_open = TRUE;
	
	my_tid = GetTID();
	
	/* Register task. */
	
	if (!Register_Task(my_object_name, my_type_name, Machine_Visible))
	{
		fprintf(stderr, "### TraceMonitor: unable to register\n");
		exit(2);
	}
	
	/* Instruct Trace Managers to begin tracing. */
	
	send_trace_managers(TM_TRACE_ON);
	
	/* Process messages. */

	while (TRUE)
	{
		SystemTask();
		
		msg = Receive(OS_MATCH_ALL, OS_MATCH_ALL, OS_MATCH_ALL, OS_NO_TIMEOUT, 0);
		if ((long) msg <= 0)
		{
			fprintf(stderr, "### TraceMonitor: Receive terminated\n");
			exit(3);
		}
		
		switch(msg->mCode)
		{
		case TM_TRACE:
			/* Copy message. */
			
			if((msg->mStatus = NetCopy(msg->mFrom, msg->mDataPtr, 
				GetTID(), &copy_msg, sizeof (struct mMessage))) == OS_NO_ERRORS)
			{
			
			/* Copy message data. */
			
				if (copy_msg.mDataSize > space)
				{
					space = copy_msg.mDataSize;
					copy_data = (char *) realloc(copy_data, space);
					if (copy_data == 0)
					{
						fprintf(stderr, 
							"### TraceMonitor: out of memory, message data too large\n");
						exit(4);
					}
				}
				
				if ((msg->mStatus = NetCopy(msg->mFrom, msg->mDataPtr + sizeof (struct mMessage),
					GetTID(), copy_data, copy_msg.mDataSize)) == OS_NO_ERRORS)
				{
				
				/* Record message. */
	
					record_message(file, &copy_msg, copy_data);
					
				}
			}
			
			/* acknowledge. */
			
			msg->mTo = msg->mFrom;
			msg->mFrom = my_tid;
			msg->mCode++;
			Send(msg);
			break;
			
		default:
			if (PREVERROR(msg->mCode))
			{
				FreeMsg(msg);
			}
			else
			{
				ERRORCODE(msg->mCode);
				
				msg->mStatus = TME_UNKNOWN_MESSAGE;
				msg->mTo = msg->mFrom;
				msg->mFrom = my_tid;
				Send(msg);
			}
			break;
		}
	}
}


/*	clean_up - Clean Up
 *
 */
 
 
static void
clean_up()
{
	void		send_trace_managers();		/* Send Message to Trace Managers */

	fprintf(stderr, "### TraceMonitor: terminating...\n");
	fflush(stderr);
	
	if (Q_open)
	{
		send_trace_managers(TM_TRACE_OFF);
		CloseQueue();
	}

	fprintf(stderr, "### TraceMonitor: terminated.\n");
	fflush(stderr);
}


/*	record_message - Record Message on File
 *
 */
 
 
void
record_message(file, msg, data)
FILE	*file;			/* Trace File Descriptor */
struct mMessage	*msg;			/* Pointer to Message */
char	*data;			/* Pointer to Message Data */
{
	int				nitems;			/* Number of Items to Write */
	int				number;			/* Number of Items Written */

	/* Record message. */
	
	nitems = sizeof (struct mMessage);
	number = fwrite(msg, sizeof(char), nitems, file);
	if (number < nitems)
	{
		fprintf(stderr, "### TraceMonitor: unable to write to Trace File\n");
		exit(6);
	}
	
	/* Record message data. */
	
	nitems = msg->mDataSize;
	number = fwrite(data, sizeof(char), nitems, file);
	if (number < nitems)
	{
		fprintf(stderr, "### TraceMonitor: unable to write to Trace File\n");
		exit(7);
	}
	
	return;
}


/*	send_trace_managers - Send Message to Trace Managers
 *
 */
 
 
void
send_trace_managers(code)
int		code;		/* Message Code to Send */
{
	int					i;									/* Temporary Index */
	unsigned short		index;								/* Name Lookup Index */
	struct mMessage		*msg;								/* Pointer to Message */
	static char			object_name[] = "Trace Manager";	/* Trace Manager Object Name */
	static char			type_name[] = "Trace Manager";		/* Trace Manager Type Name */
	struct ra_GetCards	get_cards;							/* ICC GetCards Return Area */
	tid_type			manager_tid;						/* TID of Trace Manager */
	
	/* Get TIDs of Name Managers. */
	
	while (0 == (msg = GetMsg()))
	{
		/* Nothing. */
	}
	
	msg->mTo = GetICCTID();
	msg->mFrom = GetTID();
	msg->mCode = ICC_GETCARDS;
	msg->mDataPtr = (char *) &get_cards;
	msg->mDataSize = sizeof(get_cards);
	Send(msg);
	
	msg = Receive(OS_MATCH_ALL, OS_MATCH_ALL, ICC_GETCARDS + 1, GetTickPS(), 0);
	if ((long) msg <= 0)
	{
		return;			/* ICCM timed out. */
	}
	else
	{
		FreeMsg(msg);
	}
	
	/* Lookup Trace Managers and send to them. */
	
	for (i = 0; i < IC_MAXCARDS; i++)
	{
		if (get_cards.tid[i] > 0)			/* If Name Manager present... */
		{
			index = 0;

			manager_tid = Lookup_Task(object_name, type_name, get_cards.tid[i], &index);
			
			if (manager_tid != 0)			/* If Trace Manager present... */
			{
				while (0 == (msg = GetMsg()))
				{
					/* Nothing. */
				}
				
				msg->mTo = manager_tid;
				msg->mFrom = GetTID();
				msg->mCode = code;
				Send(msg);				/* Send to Trace Manager. */
			}
		}
	}
	
	return;
}

void Wait_rcv()
{
	SystemTask();
}
